/* action.c
 *
 * Functions for storing, manipulating and executing actions
 */
#include "sim.h"
#include "server.h"

static hashtable at = {0};

/* create_action
 * Allocate a new action, and give it 4 default variables -
 * object - name of object method is associated with
 * player - name of player who activated the current method
 * method - name used by player to call current method
 * params - parameters string passed by player
 */
action *create_action(void)
{
action *na = scalloc(1,sizeof(action));

na->var[0] = create_variable("name",vt_str,vst_local);
na->var[0]->val.str = scalloc(1,1);

na->var[1] = create_variable("player",vt_str,vst_local);
na->var[1]->val.str = scalloc(1,1);

na->var[2] = create_variable("actionmethod",vt_str,vst_local);
na->var[2]->val.str = scalloc(1,1);

na->var[3] = create_variable("params",vt_strarray,vst_local);
na->var[3]->val.arr.n = 0;
na->var[3]->val.arr.a = scalloc(1,sizeof(variable));

na->varc = 4;
na->lastthought = 0;
return na;
}


/* hash table functions
 */
void add_action(action *a)
{ add_entry(&at,a); }

action *lookup_action(char *name)
{ return lookup_entry(&at,name); }

void reset_action(void)
{ reset_position(&at); }

action *next_action(void)
{ return next_position(&at); }

void remove_action(action *a)
{ remove_entry(&at,a); }

void clear_actions(void)
{ clear_entries(&at,(freefunc)free_action); }


/* free_action
 * Deep-free an action struct
 */
void free_action(action *a)
{
int i;

for(i=0; i<a->varc; i++)
	free_variable(a->var[i]);
for(i=0; i<a->metc; i++)
	free_method(a->met[i]);
free(a);
}


/* create_variable
 * Allocate and init a new variable struct
 */
variable *create_variable(char *name, int type, int storage)
{
variable *nv = scalloc(1,sizeof(variable));

nv->name = strdupe(name);
nv->type = type;
nv->storage = storage;
return nv;
}


/* free_variable
 * Deep-free a variable struct
 */
void free_variable(variable *v)
{
if (v->name)
	free(v->name);
if (v->init)
	free_pnode(v->init);
freev(v);
free(v);
}


/* freev
 * Do type-specific freeing of a variable
 */
void freev(variable *v)
{
if (v->type == vt_str)
	free(v->val.str);
if (v->type == vt_dblarray || v->type == vt_strarray) {
	int i;
	for(i=0; i<v->val.arr.n; i++)
		freev(&v->val.arr.a[i]);
	free(v->val.arr.a);
	}
}


/* create_method
 * Alloc and init a method struct
 */
method *create_method(void)
{
method *nm = scalloc(1,sizeof(method));

return nm;
}


/* free_method
 * Deep-free a method struct
 */
void free_method(method *m)
{
int i;

for(i=0; i<m->namec; i++)
	free(m->name[i]);
for(i=0; i<m->varc; i++)
	free_variable(m->var[i]);
free_pnode(m->ptree);
}


/* create_pnode
 * Alloc and init a new parse node struct
 */
pnode *create_pnode(void)
{
pnode *np = scalloc(1,sizeof(pnode));

return np;
}


/* free_pnode
 * Deep-free a parse tree node. 
 * BUG - this may not be up to date 
 */
void free_pnode(pnode *p)
{
pnode *pp, *nxt;

if (!p) return;
switch(p->type) {
case pn_if:
	free_pnode(p->val.nif.cond);
	free_pnode(p->val.nif.tr);
	free_pnode(p->val.nif.fl);
	break;
case pn_block:
	for(pp=p->val.nblock.st; pp; pp=nxt) {
		nxt = pp->next;
		free_pnode(pp);
		}
	break;
case pn_while:
	free_pnode(p->val.nwhile.cond);
	free_pnode(p->val.nwhile.blk);
	break;
case pn_for:
	free_pnode(p->val.nfor.init);
	free_pnode(p->val.nfor.cond);
	free_pnode(p->val.nfor.blk);
	free_pnode(p->val.nfor.inc);
	break;
case pn_equals:
case pn_notequals:
case pn_add:
case pn_subtract:
case pn_mult:
case pn_divide:
case pn_mod:
case pn_greater:
case pn_less:
case pn_grequals:
case pn_lsequals:
case pn_or:
case pn_and:
	free_pnode(p->val.nbinop.l);
	free_pnode(p->val.nbinop.r);
	break;
case pn_var:
	free_pnode(p->val.nvar.var);
	break;
case pn_assign:
	free_pnode(p->val.nassign.lv);
	free_pnode(p->val.nassign.val);
	break;
case pn_concat:
	free_pnode(p->val.nconcat.l);
	free_pnode(p->val.nconcat.r);
	break;
case pn_sconst:
	free(p->val.nsconst.str);
	break;
case pn_fcall:
	free(p->val.nfcall.name);
	for(pp=p->val.nfcall.ar; pp; pp=nxt) {
		nxt = pp->next;
		free_pnode(pp);
		}
	break;
case pn_aconst:
	for(pp=p->val.naconst.l; pp; pp=nxt) {
		nxt = pp->next;
		free_pnode(pp);
		}
	break;
case pn_aindex:
	free_pnode(p->val.naindex.arr);
	free_pnode(p->val.naindex.ind);
	break;
	} /* end switch */
free(p);
}


/* lookup_variable
 * Looks up a named variable in the local vars list of a method or 
 * global vars of an action.
 */
variable *lookup_variable(char *name, action *ac, method *mt)
{
int i;

if (mt) 
	for(i=0; i<mt->varc; i++)
		if (strequ(name,mt->var[i]->name))
			return mt->var[i];
if (ac)
	for(i=0; i<ac->varc; i++)
		if (strequ(name,ac->var[i]->name))
			return ac->var[i];
return NULL;
}

